package SkyNet;

use strict;
use POSIX 'strftime';
use DBI;
my $dbh;
my %usercount;

my %clients = (
	SkyNet	=> {
		nick			=> 'nunu',
		ident			=> 'nu',
		host			=> 'nu',
		ip				=> '0.0.0.0',
		cloak			=> 'nu',
		gecos			=> 'nunu',
		umode			=> '+Siopy',
		authflags	=> '$=^_><|',
		nickserv		=> 'password',
		chans			=> {
								'#acid'	=> '',
								'#a'		=> ''
							}
	}
);

my %skynet = (
	dbhost		=> 'localhost',
	dbname		=> 'database',
	dbuser		=> 'username',
	dbpass		=> 'password',
	dbtable		=> 'table',
	logchan		=> '#acid',
	trigger		=> '.',
	ucinterval	=> 60,
	ucwarning	=> 20
);

sub new {
	my ($package, %self) = @_;
	my $self = bless(\%self, $package);

	acid::setupclients(%clients);

	my $myuid = $self->{uid}{$clients{SkyNet}{nick}};

	unless ($dbh = DBI->connect("DBI:mysql:$skynet{dbname}:$skynet{dbhost}", $skynet{dbuser}, $skynet{dbpass})) {
		acid::privmsg($myuid, $skynet{logchan}, 'Failed to connect to MySQL database: %s', $DBI::errstr);
		acid::module_prepare_unload(__PACKAGE__);
		return $self;
	}

	$dbh->{mysql_auto_reconnect} = 1;

	unless ($dbh->do("CREATE TABLE IF NOT EXISTS $skynet{dbtable} (
		time TIMESTAMP,
		nick VARCHAR(30),
		ident VARCHAR(10),
		host VARCHAR(63),
		ip VARCHAR(39),
		gecos VARCHAR(50),
		umodes VARCHAR(53),
		server VARCHAR(75),
		INDEX idx_$skynet{dbtable} (nick, ip)
	)")) {
		acid::privmsg($myuid, $skynet{logchan}, 'Failed to create table: %s', $DBI::errstr);
		acid::module_prepare_unload(__PACKAGE__);
		return $self;
	}

	acid::privmsg($myuid, $skynet{logchan}, 'Connected to MySQL database %s@%s.', $skynet{dbuser}, $skynet{dbname});
	acid::privmsg($myuid, $skynet{logchan}, 'Checking for channel user count changes of %d+ every %d seconds.', $skynet{ucwarning}, $skynet{ucinterval});
	acid::privmsg($myuid, $skynet{logchan}, 'Monitoring stuff... edit me later.');

	return $self;
}

sub on_unload {
	my $self = shift;

	acid::removeclients(%clients);
	$dbh->disconnect if $dbh;
}

sub on_squit {
	my ($self, $sid, $msg) = @_;

	%usercount = (check => time + 60);
}

sub on_sid {
	my ($self, $introsid, $server, $hops, $sid, $desc) = @_;

	%usercount = (check => time + 60);
}

sub on_eob {
	my ($self, $sid) = @_;

	%usercount = (check => time + 60);
}

sub on_check {
	my $self = shift;

	my $time = time;
	return if $time - $usercount{check} < $skynet{ucinterval};

	$usercount{check} = $time;

	for (keys(%{$self->{chan}})) {
		my ($old, $new) = ($usercount{$_}, scalar(keys(%{$self->{chan}{$_}{users}})));
		my $diff = $old - $new;

		if ($old && ($diff <= "-$skynet{ucwarning}" || $diff >= $skynet{ucwarning})) {
			acid::privmsg($self->{uid}{$clients{SkyNet}{nick}}, $skynet{logchan},
				"User count of channel \002%s\002 has \002%screased\002 by \002%d\002 within the past \002%d\002 seconds. (old: \002%d\002; new: \002%d\002)",
				$_, ($old < $new ? 'in' : 'de'), abs($diff), $skynet{ucinterval}, $old, $new);
		}

		$usercount{$_} = $new;
	}
}

sub monitor {
	my ($self, $type, $chan, $uid) = @_;

	my $myuid = acid::getid($clients{SkyNet}{nick});
	
	my $info = join(' ', strftime('%Y-%m-%d %H:%M:%S', localtime), $chan, $type, acid::getname($uid), $self->{client}{$uid}{ident}, $self->{client}{$uid}{host});
	my $log = 'mod/monitored_'.$self->{conf}{monitored}{$chan}{ts}.'_'.substr($chan, 1).'.log';

	acid::privmsg($myuid, $skynet{logchan}, $info);

	unless (open LOG, '>>', $log) {
		acid::privmsg($myuid, $skynet{logchan}, 'Failed to add log entry for %s: %s', acid::getname($uid), $!);
		return;
	}
	printf LOG "%s\n", $info;
	close LOG;
}

sub on_join {
	my ($self, $uid, $ts, $chan, $mode) = @_;

	$chan = lc($chan);

	return if !$self->{conf}{monitored}{$chan};

	$self->monitor('JOIN', $chan, $uid);
}

sub on_part {
	my ($self, $uid, $chan, $msg) = @_;

	$chan = lc($chan);

	return if !$self->{conf}{monitored}{$chan};

	$self->monitor('PART', $chan, $uid);
}

sub on_uid {
	my ($self, $sid, $nick, $hops, $ts, $umode, $ident, $cloak, $ip, $uid, $host, $gecos) = @_;

	return if lc(substr($self->{server}{$sid}, 0, 4)) ne 'irc.' && lc(substr($self->{server}{$sid}, -4)) ne '.hub';

	$dbh->do("INSERT INTO $skynet{dbtable} VALUES(CURRENT_TIMESTAMP, ?, ?, ?, ?, ?, ?, ?)",
		{}, $nick, $ident, $host, $ip, $gecos, $umode, $self->{server}{$sid})
		|| acid::privmsg($uid, $skynet{logchan}, 'Failed to insert %s (%s@%s): %s', $nick, $ident, $host, $DBI::errstr);
}

sub on_privmsg {
	my ($self, $uid, $target, $msg) = @_;

	if (lc($target) eq lc($skynet{logchan})) {
		my $myuid = acid::getid($clients{SkyNet}{nick});

		if (substr($msg, 0, 1) eq $skynet{trigger}) {
			my @argv = split(' ', $msg);
			my $cmd = lc(substr(shift(@argv), 1));

			if ($cmd eq 'help') {
				acid::notice($myuid, $uid, qq(
Syntax: \002%s\037command\037 [+\037limit\037] \037query\037\002
 
\002Note:\002 The given \002\037command\037\002 must be \002NICK\002, \002IDENT\002, \002HOST\002, \002IP\002, or \002GECOS\002.
Wildcards as well as a \002+\002-prefixed limit (default is 50) are supported.
 
Examples:
 
    \002%sNICK Troublemaker\002
        Lists all records with a nick matching \002Troublemaker\002.
 
    \002%sHOST +40 *.lameisp.net\002
        Lists maximally \00240\002 records with a host matching \002*.lameisp.net\002.
 
    \002%sIP 192.168.?.*\002
        Lists all records with an IP matching \002192.168.?.*\002.
 
\002Note:\002 There also is a separate \002STATS\002 command.
 
\002Note:\002 There now also is a separate \002COHO\002 command
which takes one or more nicks and adds their most recent IP to the DNSBL.
 
\002Note:\002 There now also is a separate \002MONITOR\002 command.
),
split(//, $skynet{trigger} x 4)
				);
			}
			elsif ($cmd eq 'nick' || $cmd eq 'ip' || $cmd eq 'host' || $cmd eq 'ident' || $cmd eq 'gecos') {
				if (!@argv) {
					acid::privmsg($myuid, $target, "Syntax error, see \002%sHELP\002.", $skynet{trigger});
					return;
				}

				my ($method, $limit, $query, $count, $time) = ('=', ($argv[0] =~ /^\+\d+$/ ? shift(@argv) : 50), "@argv", 0, time);

				if ($query =~ /[*?]/) {
					$method = 'LIKE';
					$query =~ y/*?/%_/;
				}

				my $select = $dbh->selectall_arrayref("SELECT * FROM $skynet{dbtable} WHERE $cmd $method ? ORDER BY time DESC", {}, $query);

				for (@{$select}) {
					acid::privmsg($myuid, $target, '[%s] %s (%s@%s) [%s] [%s] (%s)',
						$_->[0], $_->[1], $_->[2], $_->[3], $_->[4], $_->[5], $_->[7]);
					$count++;
					last if $count == $limit;
				}
				acid::privmsg($myuid, $target, "Listed \002%d\002/\002%d\002 entries matching %s \002%s\002 in \002%s\002. (limit: \002%d\002)",
					$count, scalar(@{$select}), uc($cmd), "@argv", acid::duration(time - $time), $limit);
			}
			elsif ($cmd eq 'stats') {
				my $select = $dbh->selectrow_arrayref("SELECT
					SUBSTRING(CREATE_TIME, 1, 10),
					TABLE_ROWS,
					DATEDIFF(NOW(), CREATE_TIME),
					ROUND((DATA_LENGTH + INDEX_LENGTH) / 1024 / 1024, 0)
					FROM information_schema.TABLES WHERE TABLE_SCHEMA = ? AND TABLE_NAME = ?;",
						{}, $skynet{dbname}, $skynet{dbtable});

				acid::privmsg($myuid, $target, "Since \002%s\002 (\002%d\002 days ago) I have collected \002%d\002 entries (\002%d\002 per day) with a total size of \002%d\002 MB.",
					$select->[0], $select->[2], $select->[1], int($select->[1] / $select->[2]), $select->[3]);
			}
			elsif ($cmd eq 'coho') {
				if (!@argv) {
					acid::privmsg($myuid, $target, "Syntax error, see \002%sHELP\002.", $skynet{trigger});
					return;
				}

				for (@argv) {
					my $select = $dbh->selectrow_hashref("SELECT * FROM $skynet{dbtable} WHERE nick = ? ORDER BY time DESC LIMIT 1", {}, $_);
					if (!$select) {
						acid::privmsg($myuid, $target, '%s is not in my database.', $_);
						next;
					}

					my $hostmask = "$select->{nick}!$select->{ident}\@$select->{host}:$select->{gecos}";
					acid::privmsg($myuid, $target, 'dadding %s', $hostmask);
					acid::privmsg($myuid, '#a', '~dadd %s 5 compromised host - %s', $select->{ip}, $hostmask);
				}
			}
			elsif ($cmd eq 'monitor') {
				if (uc($argv[0]) eq 'ADD') {
					if (!$argv[1] || substr($argv[1], 0, 1) ne '#') {
						acid::privmsg($myuid, $target, "Syntax: \002%sMONITOR ADD \037#channel\037\002", $skynet{trigger});
						return;
					}

					my $chan = lc($argv[1]);

					if ($self->{conf}{monitored}{$chan}) {
						acid::privmsg($myuid, $target, "Channel \002%s\002 is already being monitored.", $chan);
						return;
					}

					$self->{conf}{monitored}{$chan}{ts} = time;
					$self->{conf}{monitored}{$chan}{nick} = acid::getname($uid);
					acid::config_save();
					acid::privmsg($myuid, $target, "Channel \002%s\002 has been added to the monitored channels.", $chan);
				}
				elsif (uc($argv[0]) eq 'DEL') {
					if (!$argv[1] || substr($argv[1], 0, 1) ne '#') {
						acid::privmsg($myuid, $target, "Syntax: \002%sMONITOR DEL \037#channel\037\002", $skynet{trigger});
						return;
					}

					my $chan = lc($argv[1]);

					if (!$self->{conf}{monitored}{$chan}) {
						acid::privmsg($myuid, $target, "Channel \002%s\002 isn't being monitored.", $chan);
						return;
					}

					delete $self->{conf}{monitored}{$chan};
					acid::config_save();
					acid::privmsg($myuid, $target, "Channel \002%s\002 has been removed from the monitored channels.", $chan);
				}
				elsif (uc($argv[0]) eq 'LIST') {
					my $count;

					for (sort { $self->{conf}{monitored}{$a}{ts} <=> $self->{conf}{$b}{ts} } keys(%{$self->{conf}{monitored}})) {
						$count++;
						acid::privmsg($myuid, $target, '%d: %s added by %s at %s', $count, $_, $self->{conf}{monitored}{$_}{nick}, scalar(localtime($self->{conf}{monitored}{$_}{ts})));
					}

					if ($count) {
						acid::privmsg($myuid, $target, 'Listed %d channels currently being monitored.', $count);
					}
					else {
						acid::privmsg($myuid, $target, 'There currently aren\'t any channels being monitored.');
					}
				}
				else {
					acid::privmsg($myuid, $target, "Syntax: \002%sMONITOR {ADD | DEL | LIST} [\037#channel\037]\002", $skynet{trigger});
				}
			}
		}
	}
}

1;
